#include<gsl/gsl_matrix.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_cblas.h>
#include <gsl/gsl_statistics_float.h> 
#include<cassert>
#include<ctime>
#include <unistd.h>
#include <cstring>
#include<algorithm>
#include<cmath>
#include <sys/time.h>  
 #include <stdio.h>  
#include <wfdb/wfdb.h>
#include<vector>
#include<string>
using namespace std;


int givenumsamples( WFDB_Siginfo *s,int sig){

  int numsamples=s[sig].nsamp*s[sig].spf;

  return numsamples;
}



void convertrecnum2str(int recnum, char*recstr){
  //recstr contains 3 chars plus'\0'

  recstr[0]='1';

  if(recnum<10){
    recstr[1]='0';
    sprintf ( recstr+2, "%d", recnum);
  } 
  else
    sprintf ( recstr+1, "%d", recnum);
}


void recnum2fname(int recordnum, const char*basedir, char*fname){

  strcpy(fname,basedir);

  char recstr[4];
  convertrecnum2str(recordnum, recstr);

  strcat(fname,recstr);
}



int loadsignalsimple(float*signal, char*fname , char* signaltype,int*baselinept,float*gainpt, int samplingfreq){

  //array signal must have alredy allocated space for the record
    

  
  int nsig;
  
 

  nsig=isigopen(fname,NULL,0);

  WFDB_Siginfo *s=new WFDB_Siginfo[nsig];

  WFDB_Sample *v=new WFDB_Sample[nsig];

  if (isigopen(fname, s, nsig) < nsig){
    printf("error reading signal\n");
    exit(1);
  }

  //setifreq(samplingfreq);

  //select channel

  int sig=findsig(signaltype);
  int numsamples=givenumsamples( s, sig);

  *baselinept=s[sig].baseline;
  *gainpt=s[sig].gain;

  if(*gainpt==0)
    *gainpt=WFDB_DEFGAIN;//default wfdb gain


  //printf(" signal is channel %d , numsamples is %d\n", sig, numsamples);
 
  int i=0;

  while(getvec(v)==nsig){

    signal[i]=v[sig];
    ++i;
  }

  assert(i==numsamples);


  return numsamples;
}


int loadsignalsetp(float*signal, int recnum , char* signaltype,
		   int*baselinept,float*gainpt, int samplingfreq){

  char basedir[]="/data/challenge2014data/set-p/";
  char fname[80];
  recnum2fname(recnum, basedir, fname);

  int length=loadsignalsimple(signal, fname , signaltype,
			      baselinept,gainpt,samplingfreq);

  return length;
}


void normalizesignal(float*signal, int numsamples){

  float mean=gsl_stats_float_mean (signal, 1, numsamples);
  float stdeviation=gsl_stats_float_sd_m (signal, 1, numsamples, mean);

  for(int i=0;i<numsamples;++i)
    signal[i]=(signal[i]-mean)/stdeviation;

  //DEBUG
  //assert(gsl_stats_float_max(signal,1,numsamples)<1.0e+2);
  //assert(gsl_stats_float_max(signal,1,numsamples)>-1.0e+2);
}



//using a vector does not require previous space allocation
//assume signal is alredy open
int loadsignalsimple(vector<float>&signal, char*fname , int sig, int nsig,
		     int samplingfreq, float begin, float end,//in secs
		     int&firstsample, int&lastsample){

  //signal does not need to be empty
  
  
  WFDB_Sample *v=new WFDB_Sample[nsig];
  WFDB_Time t, t0, t1;
 
  
  
  char buff[20];
  sprintf(buff,"%4.3f",begin);
  t0=strtim(buff);
  
  sprintf(buff,"%4.3f",end);
  t1=strtim(buff);
  
  isigsettime(t0);
 
  firstsample=t0;
  lastsample=t1;

  int i=0;

  //printf("t0=%d and t1=%d\n",t0,t1);
  
  for (t = t0; t <= t1; t++) {    
    if(getvec(v)<0)
      break;
    //signal.push_back(v[sig]);
    signal.push_back(aduphys(sig,v[sig]));

    ++i;
  }
  
  delete[] v;

  return i;//numsamples added to the signal
}






int loadannotations(char*record, char*nameannot, int fs,
		    vector<int>&annotations, int annotsamplingfreq,
		    float begin, float end//in secs)
		    ){

  //change
  //this string is now the 2nd function argument
  //char nameannot[]="atr";
  //timeintervals

  //int numannotationsatstart=annotations.size();

  WFDB_Anninfo an; 
  an.name=nameannot;
  an.stat = WFDB_READ;

  WFDB_Annotation annot;

  if(annopen(record, &an, 1)<0){
    printf("error reading annotation\n");
    exit(-2);
  }

  WFDB_Time t0, t1;

  int i=0;
  //for(int j=0;j<timeintervals.size();j+=2){

  char buff[20];
  sprintf(buff,"%4.3f",begin);
  t0=strtim(buff);
  sprintf(buff,"%4.3f",end);
  t1=strtim(buff);


  float freq=getifreq();


  iannsettime(t0*annotsamplingfreq/freq);

  while (getann(0, &annot) == 0 && annot.time <t1*annotsamplingfreq/freq){
    annotations.push_back(annot.time*freq/annotsamplingfreq-t0);
    ++i;
  }
  //lastend+=t1-t0;
  //}
 

  return i;//numnewannotations
}




void createbatchdata(const gsl_vector_float*signal, 
		     gsl_matrix_float*batchdata,
		     const int*permutation,
		     int delay,//=0 for input data, =L/2 for target data
		     int trainingstep, int batch){

  int L=batchdata->size2;
  int batchsize=batchdata->size1;

  int init=batch*batchsize, start;

  for(int i=0;i<batchsize;++i){

    start=delay+permutation[init+i]*trainingstep;

    memcpy((void*) (batchdata->data+i*L), (void*) (signal->data+start),
	   L*sizeof(float)); 
  }
}



//version using beginnings

void createbatchdata(const gsl_vector_float*signal, 
		     gsl_matrix_float*batchdata,
		     const int*permutation,
		     const vector<int>&beginnings, int delay, int batch){

  //for inputdata delayis 0
  //for output data it maybe not, for ex delay=ninputs/4 

  int L=batchdata->size2;
  int batchsize=batchdata->size1;

  int init=batch*batchsize, start;

  for(int i=0;i<batchsize;++i){

    start=beginnings[permutation[init+i]]+delay;

    memcpy((void*) (batchdata->data+i*L), (void*) (signal->data+start),
	   L*sizeof(float)); 
  }
}



void fwdnetbatch(const gsl_matrix_float*inputdata,
		 const gsl_matrix_float*w1,
		 const gsl_vector_float*bias1,
		 const gsl_matrix_float*w2,
		 const gsl_vector_float*bias2,
		 gsl_matrix_float*outputdata,
		 gsl_matrix_float*hid1data){

int batchsize=inputdata->size1;
  int numhid=hid1data->size2;
  int numoutputs=outputdata->size2;

  //insert bias in hidden units
  for(int i=0;i<batchsize;++i)
    memcpy((void*) (hid1data->data+i*numhid), (void*)(bias1->data), numhid*sizeof(float));

  gsl_blas_sgemm (CblasNoTrans, CblasNoTrans, 1.0f, inputdata, w1, 1.0, 
		  hid1data);

  //max(0,x) applied to all hidden units
  for(int i=0; i<batchsize*numhid;++i)
    *(hid1data->data+i)=std::max((float) 0.0, *(hid1data->data+i));

 //insert bias in hidden units
  for(int i=0;i<batchsize;++i)
    memcpy((void*) (outputdata->data+i*numoutputs), (void*) (bias2->data),
	   numoutputs*sizeof(float));

  gsl_blas_sgemm (CblasNoTrans, CblasNoTrans, 1.0f, hid1data, w2, 1.0, 
		  outputdata);

  //sigmoide output unit
  for(int i=0; i<batchsize*numoutputs;++i)
    *(outputdata->data+i)=1/(1+exp(-*(outputdata->data+i)));

}


void fwdnetinputsignal(const gsl_vector_float*inputsignal,
		       const gsl_matrix_float*w1, const gsl_vector_float*bias1,
		       const gsl_matrix_float*w2, const gsl_vector_float*bias2,
		       gsl_vector_float*outputsignal,
		       int batchsize, int inputpatchsize, int outputpatchsize,
		       int delay, int inputstep, int outputstep){
  //the idea is outputpatchsize=inputpatchsize/2 step=inputpatchsize/2
  //and delay=inputpatchsize/4 outputstep=outputpatchsize

  
  int ninputs=w1->size1;
  int numhid1=w1->size2;
  int numoutputs=w2->size2;

  gsl_matrix_float*inputbatchdata=gsl_matrix_float_alloc(batchsize,inputpatchsize);
  gsl_matrix_float*hid1batchdata=gsl_matrix_float_alloc(batchsize,numhid1);
  gsl_matrix_float*outputbatchdata=gsl_matrix_float_alloc(batchsize,outputpatchsize);

  int numsamples=inputsignal->size;
  int numbatches=numsamples/(batchsize*inputstep);

  int idpermut[numsamples];
  for(int i=0;i<numsamples;++i)
    idpermut[i]=i;



  for(int i=0;i<numbatches;++i){

    createbatchdata(inputsignal, inputbatchdata, idpermut, 0, inputstep, i);

    fwdnetbatch(inputbatchdata, w1, bias1, w2, bias2, outputbatchdata,
		hid1batchdata);

    memcpy((void*) (outputsignal->data+delay+i*batchsize*outputstep),
	   (void*) (outputbatchdata->data), batchsize*outputpatchsize*sizeof(float));
  }


  //its missing from the signal what is left after the last batch
  //int left=numsamples-delay-numbatches*batchsize*outputpatchsize;
  int start=numbatches*batchsize*numoutputs;
  int end=start+ninputs-1;

  while(end<numsamples){
    //use only one raw from batchdata matrix
    memcpy((void*) inputbatchdata->data, (void*) (inputsignal->data+start),
	   ninputs*sizeof(float));
    fwdnetbatch(inputbatchdata, w1, bias1, w2, bias2, outputbatchdata,
		hid1batchdata);

    memcpy((void*) (outputsignal->data+delay+start),
	   (void*) outputbatchdata->data, numoutputs*sizeof(float));//only first raw
    start+=numoutputs;
    end+=numoutputs;
  }

  gsl_matrix_float_free(outputbatchdata);
  gsl_matrix_float_free(hid1batchdata);
  gsl_matrix_float_free(inputbatchdata);

}






//passes a batch trought the net
void fwdautoencoderbatch(const gsl_matrix_float*inputdata,
			 const gsl_matrix_float*w,
			 const gsl_vector_float*bias1,
			 const gsl_vector_float*bias2,
			 gsl_matrix_float*outputdata,
			 gsl_matrix_float*codedata){

  int batchsize=inputdata->size1;
  int codeL=codedata->size2;
  int ninputs=outputdata->size2;

  //insert bias in hidden units
  for(int i=0;i<batchsize;++i)
    memcpy((void*) (codedata->data+i*codeL), (void*)(bias1->data), codeL*sizeof(float));

  gsl_blas_sgemm (CblasNoTrans, CblasNoTrans, 1.0f, inputdata, w, 1.0, 
		  codedata);

  //max(0,x) applied to all hidden units
  for(int i=0; i<batchsize*codeL;++i)
    *(codedata->data+i)=std::max((float) 0.0, *(codedata->data+i));

 //insert bias in hidden units
  for(int i=0;i<batchsize;++i)
    memcpy((void*) (outputdata->data+i*ninputs), (void*) (bias2->data),
	   ninputs*sizeof(float));

  gsl_blas_sgemm (CblasNoTrans, CblasTrans, 1.0f, codedata, w, 1.0, 
		  outputdata);

  //output units are id
}




//divides a signal in batches and buils a long signal from the outputbatches
void fwdautoencoderinputsignal(const gsl_vector_float*inputsignal,
		       const gsl_matrix_float*w, const gsl_vector_float*bias1,
		       const gsl_vector_float*bias2,
		       gsl_vector_float*outputsignal,
		       int batchsize){
  //the idea is outputpatchsize=inputpatchsize/2 step=inputpatchsize/2
  //and delay=inputpatchsize/4 outputstep=outputpatchsize

  
  int codeL=w->size2;
  int ninputs=w->size1;
  int inputstep=ninputs;

  gsl_matrix_float*inputbatchdata=gsl_matrix_float_alloc(batchsize,ninputs);
  gsl_matrix_float*codebatchdata=gsl_matrix_float_alloc(batchsize,codeL);
  gsl_matrix_float*outputbatchdata=gsl_matrix_float_alloc(batchsize,ninputs);

  int numsamples=inputsignal->size;
  int numbatches=numsamples/(batchsize*inputstep);

  int idpermut[numsamples];
  for(int i=0;i<numsamples;++i)
    idpermut[i]=i;

  for(int i=0;i<numbatches;++i){

    createbatchdata(inputsignal, inputbatchdata, idpermut, 0, inputstep, i);

    fwdautoencoderbatch(inputbatchdata, w, bias1, bias2, outputbatchdata,
		codebatchdata);

    memcpy((void*) (outputsignal->data+i*batchsize*ninputs),
	   (void*) (outputbatchdata->data), batchsize*ninputs*sizeof(float));
  }


  //its missing from the signal what is left after the last batch->TODO

  gsl_matrix_float_free(outputbatchdata);
  gsl_matrix_float_free(codebatchdata);
  gsl_matrix_float_free(inputbatchdata);

}



//in this version we
//use only the middle of the intervals to build the long outputsignal
void fwdautoencoderinputsignal(const gsl_vector_float*inputsignal,
			       const gsl_matrix_float*w,
			       const gsl_vector_float*bias1,
			       const gsl_vector_float*bias2,
			       gsl_vector_float*outputsignal,
			       int batchsize,int outputL){

  //we use only the middle outputL samples from the output of each
  // input interval
  
  int codeL=w->size2;
  int ninputs=w->size1;
 
  int delay=(ninputs-outputL)/2;

  int inputstep=outputL;

  gsl_matrix_float*inputbatchdata=gsl_matrix_float_alloc(batchsize,ninputs);
  gsl_matrix_float*codebatchdata=gsl_matrix_float_alloc(batchsize,codeL);
  gsl_matrix_float*outputbatchdata=gsl_matrix_float_alloc(batchsize,ninputs);

  int numsamples=inputsignal->size;
  int numbatches=numsamples/(batchsize*inputstep);

  int numbeginnings=numsamples/inputstep;
  int idpermut[numbeginnings];
  for(int i=0;i<numbeginnings;++i)
    idpermut[i]=i;

  for(int i=0;i<numbatches;++i){

    createbatchdata(inputsignal, inputbatchdata, idpermut, 0, inputstep, i);

    fwdautoencoderbatch(inputbatchdata, w, bias1, bias2, outputbatchdata,
		codebatchdata);

    for(int j=0;j<batchsize;++j)
      memcpy((void*) (outputsignal->data+delay+(i*batchsize+j)*outputL),
	   (void*) (outputbatchdata->data+j*ninputs+delay), outputL*sizeof(float));
  }

  // in the beginning there is a small piece  missing,
  // corresponding to the delay
  memcpy((void*) inputbatchdata->data, (void*) inputsignal->data,
	 ninputs*sizeof(float));
  fwdautoencoderbatch(inputbatchdata, w, bias1, bias2, outputbatchdata,
		      codebatchdata);
  memcpy((void*) outputsignal->data,
	 (void*) outputbatchdata->data, delay*sizeof(float));


  //its missing from the signal what is left after the last batch
  int left=numsamples-delay-numbatches*batchsize*outputL;

  int start=numbatches*batchsize*outputL;
  int end=start+ninputs-1;

  while(end<numsamples){
    //use only one raw from batchdata matrix
    memcpy((void*) inputbatchdata->data, (void*) (inputsignal->data+start),
	   ninputs*sizeof(float));
    fwdautoencoderbatch(inputbatchdata, w, bias1, bias2, outputbatchdata,
			codebatchdata);
    memcpy((void*) (outputsignal->data+delay+start),
	   (void*) (outputbatchdata->data+delay), outputL*sizeof(float));
    start+=outputL;
    end+=outputL;
  }
		  
  //there might be still a small piece of output signal left
 
  int numpoints=numsamples-end+delay+outputL;
  if(numpoints>0){
    start=numsamples-ninputs;
    memcpy((void*) inputbatchdata->data, (void*) (inputsignal->data+start),
	   ninputs*sizeof(float));
    fwdautoencoderbatch(inputbatchdata, w, bias1, bias2, outputbatchdata,
			codebatchdata);
    memcpy((void*) (outputsignal->data+numsamples-numpoints),
	   (void*) (outputbatchdata->data+delay), numpoints*sizeof(float));
  }

  gsl_matrix_float_free(outputbatchdata);
  gsl_matrix_float_free(codebatchdata);
  gsl_matrix_float_free(inputbatchdata);

}












void savebiasesandweightsautoencoder(const char*fname, 
				     const gsl_matrix_float *w,
				     const gsl_vector_float*biases1,
				     const gsl_vector_float*biases2){

  FILE*fid=fopen(fname,"w");

  //w
  fprintf(fid,"%d\n",(int) w->size1);
  fprintf(fid,"%d\n",(int) w->size2);

  int numelements=(w->size1)*(w->size2);

  for(int i=0;i<numelements;++i)
    fprintf(fid,"%f\n",*(w->data+i));

  //biases1
  numelements=biases1->size;
  fprintf(fid,"%d\n",numelements);
  for(int i=0;i<numelements;++i)
    fprintf(fid,"%f\n",*(biases1->data+i));

  //biases2
  numelements=biases2->size;
  fprintf(fid,"%d\n",numelements);
  for(int i=0;i<numelements;++i)
    fprintf(fid,"%f\n",*(biases2->data+i));

  fclose(fid);
}



void loadbiasesandweightsautoencoder(const char*fname, 
				     gsl_matrix_float *w,
				     gsl_vector_float*biases1,
				     gsl_vector_float*biases2){


  FILE*fid=fopen(fname,"r");

  int numelements;

  //w
  fscanf(fid,"%d",&numelements);
  assert(numelements==(int) w->size1);

  fscanf(fid,"%d",&numelements);
  assert(numelements==(int) w->size2);

  numelements=(w->size1)*(w->size2);

  for(int i=0;i<numelements;++i)
    fscanf(fid,"%f",w->data+i);

  //biases1
  fscanf(fid,"%d",&numelements);
  assert(numelements==(int) biases1->size);
  for(int i=0;i<numelements;++i)
    fscanf(fid,"%f",biases1->data+i);

  //biases2
  fscanf(fid,"%d",&numelements);
  assert(numelements==(int) biases2->size);
  for(int i=0;i<numelements;++i)
    fscanf(fid,"%f",biases2->data+i);

  fclose(fid);
}



void sum_matrix_columns( gsl_vector_float*v, const gsl_matrix_float*mat){

  int numrows=mat->size1;
  int numcols=mat->size2;

  memset((void*)v->data, 0, numcols*sizeof(float));

  for (int i=0;i<numcols;++i)
    for(int j=0;j<numrows;++j)
      *(v->data+i)+=*(mat->data+j*numcols+i);
}



//logistic  units in the last layer
float givesamplerrorsigmoide(const gsl_matrix_float*inputdata,
			     const gsl_matrix_float*targetdata,
			     gsl_matrix_float*hid1data,
			     gsl_matrix_float*outputdata,
			     const gsl_matrix_float*w1,
			     const gsl_vector_float*bias1,
			     const gsl_matrix_float*w2,
			     const gsl_vector_float*bias2){

  int batchsize=inputdata->size1;
  int numhid=hid1data->size2;
  int numoutputs=outputdata->size2;

  //insert bias in hidden units
  for(int i=0;i<batchsize;++i)
    memcpy((void*) (hid1data->data+i*numhid), (void*)(bias1->data), numhid*sizeof(float));

  gsl_blas_sgemm (CblasNoTrans, CblasNoTrans, 1.0f, inputdata, w1, 1.0, 
		  hid1data);

  //max(0,x) applied to all hidden units
  for(int i=0; i<batchsize*numhid;++i)
    *(hid1data->data+i)=std::max((float) 0.0, *(hid1data->data+i));

 //insert bias in hidden units
  for(int i=0;i<batchsize;++i)
    memcpy((void*) (outputdata->data+i*numoutputs), (void*) (bias2->data),
	   numoutputs*sizeof(float));

  gsl_blas_sgemm (CblasNoTrans, CblasNoTrans, 1.0f, hid1data, w2, 1.0, 
		  outputdata);

  //sigmoide output unit
  for(int i=0; i<batchsize*numoutputs;++i)
    *(outputdata->data+i)=1/(1+exp(-*(outputdata->data+i)));

  //assert(gsl_matrix_float_max(outputdata)<2);

  gsl_matrix_float_sub(outputdata,targetdata);

  gsl_matrix_float_scale(outputdata,-1.0);//outputdata=errdata

  gsl_vector_float_const_view auxvec=
    gsl_vector_float_const_view_array(outputdata->data, numoutputs*batchsize);

  float avgerr=gsl_blas_snrm2(&auxvec.vector);

  return avgerr*avgerr/batchsize;
}


void savebiasesandweightsnnet1hid(const char*fname, 
				     const gsl_matrix_float *w1,
				     const gsl_matrix_float *w2,
				     const gsl_vector_float*biases1,
				     const gsl_vector_float*biases2){

  FILE*fid=fopen(fname,"w");

  //w1
  fprintf(fid,"%d\n",(int) w1->size1);
  fprintf(fid,"%d\n",(int) w1->size2);

  int numelements=(w1->size1)*(w1->size2);

  for(int i=0;i<numelements;++i)
    fprintf(fid,"%f\n",*(w1->data+i));


  //w2
  fprintf(fid,"%d\n",(int) w2->size1);
  fprintf(fid,"%d\n",(int) w2->size2);

  numelements=(w2->size1)*(w2->size2);

  for(int i=0;i<numelements;++i)
    fprintf(fid,"%f\n",*(w2->data+i));

  //biases1
  numelements=biases1->size;
  fprintf(fid,"%d\n",numelements);
  for(int i=0;i<numelements;++i)
    fprintf(fid,"%f\n",*(biases1->data+i));

  //biases2
  numelements=biases2->size;
  fprintf(fid,"%d\n",numelements);
  for(int i=0;i<numelements;++i)
    fprintf(fid,"%f\n",*(biases2->data+i));

  fclose(fid);
}



void loadbiasesandweightsnnet1hid(const char*fname, 
				     gsl_matrix_float *w1,
				     gsl_matrix_float *w2,
				     gsl_vector_float*biases1,
				     gsl_vector_float*biases2){


  FILE*fid=fopen(fname,"r");

  int numelements;

  //w1
  fscanf(fid,"%d",&numelements);
  assert(numelements==(int) w1->size1);

  fscanf(fid,"%d",&numelements);
  assert(numelements==(int) w1->size2);

  numelements=(w1->size1)*(w1->size2);

  for(int i=0;i<numelements;++i)
    fscanf(fid,"%f",w1->data+i);

 //w2
  fscanf(fid,"%d",&numelements);
  assert(numelements==(int) w2->size1);

  fscanf(fid,"%d",&numelements);
  assert(numelements==(int) w2->size2);

  numelements=(w2->size1)*(w2->size2);

  for(int i=0;i<numelements;++i)
    fscanf(fid,"%f",w2->data+i);

  //biases1
  fscanf(fid,"%d",&numelements);
  assert(numelements==(int) biases1->size);
  for(int i=0;i<numelements;++i)
    fscanf(fid,"%f",biases1->data+i);

  //biases2
  fscanf(fid,"%d",&numelements);
  assert(numelements==(int) biases2->size);
  for(int i=0;i<numelements;++i)
    fscanf(fid,"%f",biases2->data+i);

  fclose(fid);
}


//create a signal with a peak for each annotation
void createsignalqrs(vector<int>&annotations,float*signalqrs,
		       int samplingfreq, int numsamples, int numannot){

  int qrsradius=150*samplingfreq/1000;//150 ms

  int sample;

  memset((void*) signalqrs, 0, numsamples*sizeof(float));

  for(int i=0;i<numannot;++i){

    signalqrs[annotations[i]]=1;

    for(int j=1;j<qrsradius;++j){

      sample=annotations[i]+j;
      if(sample<numsamples)
	signalqrs[sample]=(float)(1.0*qrsradius-j)/qrsradius;


      sample=annotations[i]-j;
      if(sample>=0)
	signalqrs[sample]=(float)(1.0*qrsradius-j)/qrsradius;
    }

  }
}


