#include <stdio.h>
#include <stdlib.h>
#include "beatlib.h"

Beat_array *Beat_array_init ( int nelements){
  
  Beat_array *out = malloc(sizeof(Beat_array)) ;
  
  if ( out == NULL ) 
    return NULL ;
  
  out->number = nelements ;
  out->fs = 0 ;
  out->mem = calloc( nelements , sizeof(int)) ;

  if ( out->mem == NULL )
    return NULL ;
  
  return out ;
}

void Beat_array_free( Beat_array *in ){
  if ( in == NULL )
    return ;
  
  free(in->mem ) ;
  free(in) ;
  return ;
}

/* usually all information is placed on files */
Beat_array* Beat_array_from_File( char *filename ){

  FILE *qrsfile = fopen(filename , "r") ;
  int qrsnumber ;
  fscanf(qrsfile , "%d" , &qrsnumber ) ;

  Beat_array *out = Beat_array_init(qrsnumber) ;
  fscanf(qrsfile , "%d" , &out->fs ) ;
  
  int i ;
  for ( i = 0 ; i < out->number ; i++ )
    fscanf(qrsfile , "%d" , &out->mem[i]) ;

  fclose(qrsfile) ;
  return out ;

}


PBeat_array *PBeat_array_init ( int number ){

  PBeat_array *out = malloc(sizeof(PBeat_array)) ;
  out->number = number ;
  out->original= calloc(number , sizeof(int));
  out->parallel = calloc(number , sizeof(int)) ;
  return out ;
}

void PBeat_array_free(PBeat_array *in){
  
  if (in == NULL )
    return ;
  
  free(in->original) ;
  free(in->parallel) ;
  free(in) ;

  return ;
}

PBeat_array *PBeat_from_arrays(Beat_array *abeat , Beat_array *bbeat){
  
  PBeat_array *out = PBeat_array_init(abeat->number) ;

  int i , j = 0 ;
  for ( i = 0 ; i < abeat->number ; i++ ){
    out->original[i]=i ;
    if ( j < bbeat->number - 1 ){
      while ( abs(abeat->mem[i] - bbeat->mem[j]) > abs(abeat->mem[i] - bbeat->mem[j+1]) && j < bbeat->number - 1)
	j++ ;
      out->parallel[i] = j ;
    }
    else 
      out->parallel[i] = bbeat->number - 1 ;
  }

  return out ;
}


IBeat_array *IBeat_array_init(int number){
  
  IBeat_array *out = malloc(sizeof(IBeat_array)) ;
  out->number = number ;
  out->mem = calloc(number , sizeof(int)) ;
  out->beg = calloc(number , sizeof(int)) ;
  out->end = calloc(number , sizeof(int)) ;
  return out ;
}

void IBeat_array_free(IBeat_array *in){
 
  if (in == NULL)
    return ;
 
  free(in->mem) ;
  free(in->beg) ;
  free(in->end) ;
  free(in) ;
  
  return ;
}

IBeat_array *NIBeat_from_array(Beat_array *in , /* tipically this is a list of qrs by some detector */
			      int fs , /* the sampling frequency of the record */
			      int length ){ /* this is the length of the interval where we are going to select beats */

 
  int jump = fs * length / 2 ;

  /* we need to find the first and the last beat detected from which we can spread the interval (-jump,jump) */
  int jinit = 0 ;
  while ( abs(in->mem[0] - in->mem[jinit]) < jump )
    jinit++ ;
  
  int jfinal = in->number - 1 ;
  while ( abs(in->mem[in->number - 1] - in->mem[jfinal]) < jump )
    jfinal-- ;

  /* printf("%d %d\n" , jinit , jfinal ) ; */

  IBeat_array *out = IBeat_array_init(in->number) ;

  int i , j , k ;
  k = 0 ;
  
  /* we just anotate the postion of the beats and not its time ocorrence*/
  for ( i = 0 ; i < jinit ; i++ ){
    out->mem[i] = i ;
    out->beg[i] = 0 ;
    while ( abs(in->mem[i] - in->mem[k] ) < jump)
      k++ ;
    out->end[i] = k ;
    /* printf("begin: %d < position: %d < end: %d\n" , out->beg[i] , out->mem[i] , out->end[i] ) ; */
  }
  
  j = 0 ;
  k = jinit ;

  for ( i = jinit ; i < in->number ; i++ ){
    out->mem[i] = i ;
    while ( abs(in->mem[i] - in->mem[j])> jump )
      j++ ;
    out->beg[i] = j ;
    while ( abs( in->mem[i] - in->mem[k] ) < jump && k < in->number)
      k++ ;
    out->end[i] = k ;
    /* printf("begin: %d < position: %d < end: %d\n" , out->beg[i] , out->mem[i] , out->end[i] ) ; */
  }
  
  return out ;
}

/* this is the old version of the function. the new version considers all beats in the file. */
IBeat_array *IBeat_from_array(Beat_array *in , /* tipically this is a list of qrs by some detector */
			      int fs , /* the sampling frequency of the record */
			      int length ){ /* this is the length of the interval where we are going to select beats */

 
  int jump = fs * length / 2 ;

  /* we need to find the first and the last beat detected from which we can spread the interval (-jump,jump) */
  int jinit = 0 ;
  while ( abs(in->mem[0] - in->mem[jinit]) < jump )
    jinit++ ;
  /* printf("%d\n" , jinit ) ; */
  
  int jfinal = 0 ;
  while ( abs(in->mem[in->number - 1] - in->mem[in->number - 1 - jfinal]) < jump )
    jfinal++ ;
  /* printf("%d\n" , jfinal ) ; */

  /* ok! from the total number of beats the number that have a spread interval is:*/
  int number_total = in->number - jfinal - jinit ;
  
  IBeat_array *out = IBeat_array_init(number_total) ;

  int i , j , k ;
  i = 0 ;
  j = 0 ;
  k = jinit + i ;
  
  for ( i = 0 ; i < out->number ; i++ ){
    out->mem[i] = jinit + i ;
    while ( abs(in->mem[jinit + i] - in->mem[j])> jump && j < in->number - jfinal -1 )
      j++ ;
    out->beg[i] = j ;
    while ( abs(in->mem[jinit + i] - in->mem[k] ) <jump && k < in->number - 1)
      k++ ;
    out->end[i] = k ;
  }
    
  return out ;
}


int soma (int *lista , int initl , int number ){
  int out = 0 ;
  int i ; 
  for ( i = initl ; i < initl + number ; i++ )
    out += lista[i] ;
  return out ;
}

QBeat_array *QBeat_array_init(int number){

  QBeat_array *out = malloc(sizeof(QBeat_array)) ;
  out->number = number ;
  out->mem = calloc(number , sizeof(int)) ;
  out->quality = calloc(number , sizeof(float)) ;
  return out ;
}

void QBeat_array_free(QBeat_array *in){
  
  free(in->mem) ;
  free(in->quality) ;
  free(in) ;

  return ;
}

/* /\* this is the original version. it doesn't count with both tails of the signal*\/ */
/* QBeat_array *QBeat_from_arrays(Beat_array *a_beat , Beat_array *b_beat , int fs , int length ){ */

/*   int i ; */
/*   PBeat_array *bijection = PBeat_from_arrays(a_beat , b_beat) ; */

/*   IBeat_array *ia_beat = IBeat_from_array(a_beat , fs , length ) ; */
/*   IBeat_array *ib_beat = IBeat_from_array(b_beat , fs , length ) ; */

/*   int *NA = calloc(ia_beat->number , sizeof(int)) ; */
/*   int *NB = calloc(ia_beat->number , sizeof(int)) ; */
  
/*   for ( i = 0 ; i < ia_beat->number ; i++ ){ */
/*     NA[i] = ia_beat->end[i] - ia_beat->beg[i] + 1 ; */
/*     NB[i] = ib_beat->end[bijection->parallel[i]] - ib_beat->beg[bijection->parallel[i]] + 1 ; */
/*   } */

/*   float error = (TOLERANCE * fs) / 1000 ; */

/*   int initl ; */
/*   if ( ia_beat->beg[0] < ib_beat->beg[0] ) */
/*     initl = ia_beat->beg[0] ; */
/*   else */
/*     initl = ib_beat->beg[0] ; */

/*   int final ; */
/*   if ( ia_beat->end[ia_beat->number - 1] < ib_beat->end[ib_beat->number - 1] ) */
/*     final = ib_beat->end[ib_beat->number - 1] ; */
/*   else */
/*     final = ia_beat->end[ia_beat->number - 1] ; */

/*   int diff_num = final - initl + 1 ; */

/*   /\***********************************************************************************************************\/ */
/*   int *NMlocal = calloc(diff_num , sizeof(int)) ; */
/*   int tmp_init = 0 ; */
/*   for ( i = 0 ; i < bijection->number ; i++ ) */
/*     if (abs(a_beat->mem[bijection->original[i]] - b_beat->mem[bijection->parallel[i]]) < error ) */
/*       NMlocal[i] = 1 ; */

/*   /\* tmp_init = ia_beat->mem[0] ; *\/ */
/*   /\* tmp_end = ia_beat->mem[0] + ia_beat->number ; *\/ */
/*   /\* for ( i = tmp_init ; i < tmp_end ; i++ ) *\/ */
/*   /\*   if (abs(a_beat->mem[ia_beat->mem[i]] - b_beat->mem[ib_beat->mem[bijection->parallel[i]]]) < error ) *\/ */
/*   /\*     NMlocal[i] = 1 ; *\/ */

/*   /\* tmp_init = ia_beat->mem[0] + ia_beat->number ; *\/ */
/*   /\* tmp_end = diff_num ; *\/ */
/*   /\* for ( i = tmp_init ; i < tmp_end ; i++ ) *\/ */
/*   /\*   if (abs(a_beat->mem[ia_beat->end[i]] - b_beat->mem[ib_beat->end[bijection->parallel[i]]]) < error ) *\/ */
/*   /\*     NMlocal[i] = 1 ; *\/ */
  
/*   /\***********************************************************************************************************\/ */

/*   int *NM = calloc(ia_beat->number , sizeof(int)) ; */
/*   for ( i = 0 ; i < ia_beat->number ; i++ ){ */
/*     tmp_init = ia_beat->beg[i] ; */
/*     NM[i] = soma(NMlocal , tmp_init , NA[i] ) ; */
/*   } */
/*   printf("\n") ; */

/*   QBeat_array *out = QBeat_array_init(ia_beat->number) ; */
/*   for ( i = 0 ; i < out->number ; i++ ){ */
/*     out->mem[i] = ia_beat->mem[i] ; */
/*     out->quality[i] = (NM[i] + 0.0)/(NA[i] + NB[i] - NM[i]) ; */
/*   } */

/*   free(NA) ; */
/*   free(NB) ; */
/*   free(NMlocal) ; */
/*   free(NM) ; */
/*   PBeat_array_free(bijection) ; */
/*   IBeat_array_free(ia_beat) ; */
/*   return out ; */

/* } */

int min_int( int a , int b ){
  
  int out ;
  if ( a < b )
    out = a ;
  else 
    out = b ;

  return out ;
}


void timeformat( int position , int fs , char *tempo){
  int sec = position / fs ;
  
  int segundos , minutos , horas ;
  segundos = sec % 60 ;
  minutos = (sec / 60) % 60 ;
  horas = (sec / 3600) % 60 ;
  
  if (minutos < 10 ){
    if (segundos < 10 )
      sprintf(tempo , "%d:0%d:0%d" , horas , minutos , segundos ) ;
    else
      sprintf(tempo , "%d:0%d:%d" , horas , minutos , segundos ) ;
  }
  else {
    if (segundos < 10 )
      sprintf(tempo , "%d:%d:0%d" , horas , minutos , segundos ) ;
    else 
      sprintf(tempo , "%d:%d:%d" , horas , minutos , segundos ) ;
  }
  return ;
}

QBeat_array *NQBeat_from_arrays(Beat_array *a_beat , Beat_array *b_beat , int fs , int length ){

  int i ;
  /* we start by finding the correspondent (parallel) beats in the two signals */
  PBeat_array *bijection = PBeat_from_arrays(a_beat , b_beat) ;

  /* after we spread around each position the beat positions that are inside the defined interval */
  IBeat_array *ia_beat = NIBeat_from_array(a_beat , fs , length ) ;
  IBeat_array *ib_beat = NIBeat_from_array(b_beat , fs , length ) ;

  /* these arrays just keep the number of beats inside each interval for each array */
  int *NA = calloc(bijection->number , sizeof(int)) ;
  int *NB = calloc(bijection->number , sizeof(int)) ;
  
  for ( i = 0 ; i < ia_beat->number ; i++ ){
    NA[i] = ia_beat->end[bijection->original[i]] - ia_beat->beg[bijection->original[i]] + 1 ;
    NB[i] = ib_beat->end[bijection->parallel[i]] - ib_beat->beg[bijection->parallel[i]] + 1 ;
  }

  /* for ( i = 0 ; i < bijection->number ; i++ ) */
  /*   printf("begin %d < original %d < end %d (NA = %d) -> beging %d < parallel %d < end %d (NB = %d)\n" ,  */
  /* 	   ia_beat->beg[bijection->original[i]] , ia_beat->mem[bijection->original[i]] , ia_beat->end[bijection->original[i]] , */
  /* 	   NA[i] , */
  /* 	   ib_beat->beg[bijection->parallel[i]] , ib_beat->mem[bijection->parallel[i]] , ib_beat->end[bijection->parallel[i]] , NB[i]) ; */


  float error = (TOLERANCE * fs) / 1000 ;

  /***********************************************************************************************************/
  
  int *NMlocal = calloc(bijection->number , sizeof(int)) ;

  for ( i = 0 ; i < bijection->number ; i++ )
    if (abs(a_beat->mem[bijection->original[i]] - b_beat->mem[bijection->parallel[i]] ) < error )
      NMlocal[bijection->original[i]] = 1 ;
  
  /***********************************************************************************************************/
  int tmp_init_a = 0 ;
  int tmp_init_b = 0 ;
  int tmp_lixo_a ;
  int tmp_lixo_b ;

  int *NM = calloc(bijection->number , sizeof(int)) ;
  for ( i = 0 ; i < bijection->number ; i++ ){
    tmp_init_a = ia_beat->beg[bijection->original[i]] ;
    tmp_init_b = ib_beat->beg[bijection->parallel[i]] ;
    tmp_lixo_a = soma(NMlocal , tmp_init_a , NA[i]) ;
    tmp_lixo_b = soma(NMlocal , tmp_init_b , NB[i]) ;
    printf("position : %d init->a %d , init->b %d, soma->a %d , soma->b %d\n" , i , tmp_init_a , tmp_init_b , tmp_lixo_a , tmp_lixo_b ) ;
    NM[i] = min_int(soma(NMlocal , tmp_init_a , NA[i]) , soma(NMlocal , tmp_init_b , NB[i] )) ; 
  }

  char tempo[20] ;

  for ( i = 0 ; i < bijection->number ; i++ ){
    timeformat(a_beat->mem[bijection->original[i]], fs , tempo ) ;
    printf("%d %s ---> begin %d < original %d < end %d (NA = %d) -> beging %d < parallel %d < end %d (NB = %d) :: %d :: %d\n" ,
  	   a_beat->mem[bijection->original[i]] , tempo ,
  	   ia_beat->beg[bijection->original[i]] , ia_beat->mem[bijection->original[i]] , ia_beat->end[bijection->original[i]] ,
  	   NA[i] ,
  	   ib_beat->beg[bijection->parallel[i]] , ib_beat->mem[bijection->parallel[i]] , ib_beat->end[bijection->parallel[i]] ,
  	   NB[i] , NMlocal[i] , NM[i] ) ;
  }

  printf("\n") ;

  QBeat_array *out = QBeat_array_init(bijection->number) ;
  for ( i = 0 ; i < bijection->number ; i++ ){
    out->mem[i] = ia_beat->mem[bijection->original[i]] ;
    out->quality[i] = (NM[i] + 0.0)/(NA[i] + NB[i] - NM[i]) ;
  }

  free(NA) ;
  free(NB) ;
  free(NMlocal) ;
  free(NM) ;
  PBeat_array_free(bijection) ;
  IBeat_array_free(ia_beat) ;
  return out ;

}

