#include <cstdlib>
#include <cstdio>
#include <cassert>
#include <cstring>
#include <vector>
#include <algorithm>
using namespace std;



void selectsignalsonintervals(vector<int>&ecgsignals,vector<int>&bpsignals,
			      vector<vector<float> >&ecgspower,
			      vector<vector<float> >&ecgscomparedetections,
			      vector<vector<float> >&bpspower,
			      vector<vector<float> >&bpscomparedetections,
			      vector<int>&signalonintervals,
			      float ecgspowerlimit,float ecgscomparedetectionslimit,
			      float bpspowerlimit,float bpscomparedetectionslimit);


//read ecg(s) and bp(s) quality indices, select (if possible) one signal for each interval


//list, for each used signal, on which intervals it is the "chosen one"
void listintervalsforsignals(const vector<int>&signalonintervals, vector<int>&usedsignals,
			     vector<vector<int> >& intervalsforsignals);

int selectannotationsonintervalstxt(char*rec,char*annotname,//rec includes signal num
				    vector<int>&intervals,  int lengthsecs,
				    vector<vector<int> >&detectionsonintervals);

void createfinalannotations(const vector<int>&signalonintervals, const vector<int>&usedsignals,
			    vector<vector<vector<int> > >&detectionsonintervals,
			    vector<int>&finalannot);


int main(int argc,char**argv){


  //arguments are recordpath rec  samplingfreq and signals numbers (first ECG and then BP)
  // mgh/mgh001 mgh001 250 0 3 1 2 ECG ECG BP BP

  int lengthsecs=5;//part of interval length in secs used to extrract annotations

  //this values still need to be determined
  float ecgspowerlimit=0.005;
  float ecgscomparedetectionslimit=0.7;
  float bpspowerlimit=0.1;
  float bpscomparedetectionslimit=0;//not in use

  int numsignals=(argc-4)/2;
  assert(2*numsignals+4==argc);


  vector<int> ecgsignals,bpsignals;//signal numbers corresponding to ecgs and bps
  //determine  ecg(s) and bp(s) 
  int numecgs=0, numbps=0, j=0;
  while(j<numsignals){

    if(strcmp(argv[j+4+numsignals],"ECG")==0){
      ecgsignals.push_back(atoi(argv[j+4]));
      numecgs++;
    }
    else 
      if(strcmp(argv[j+4+numsignals],"bloodpressure")==0){
	bpsignals.push_back(atoi(argv[j+4]));
	numbps++;
      }
      else{
	printf("wrong type of signal %d %s\n",atoi(argv[j+4]),argv[j+4+numsignals]);
	exit(-1);
      }
    j++;
  }
 
  //debug
  //
  //ecgs
  printf("there are %d ECG signals:\n",numecgs);
  for(int i=0;i<numecgs;++i)
    printf("%d \t",ecgsignals[i]);
  printf("\n");
  //bps
  printf("there are %d BP signals:\n",numbps);
  for(int i=0;i<numbps;++i)
    printf("%d \t",bpsignals[i]);
  printf("\n");


 
  char fname[50];
  float value;
  char numstr[21];
  FILE*fid;

  //read ECGs quality indices int vectors
  //
  vector<vector<float> >ecgspower(numecgs);
  vector<vector<float> >ecgscomparedetections(numecgs);

  for(int i=0;i<numecgs;++i){

    //ecgpower  
    //
    strcpy(fname,"powersignalnobase");
    sprintf(numstr, "%d", ecgsignals[i]);
    strcat(fname,numstr);
    strcat(fname,".txt");
    fid=fopen(fname,"r");
    if(fid==NULL){
      printf("can not open file %s\n",fname);fflush(stdout);
      exit(-1);
    }
    while(fscanf(fid,"%f",&value)!=EOF)
      ecgspower[i].push_back(value);
    fclose(fid);


    //ecgscomparedetections between wqrs and gqrs
    //
    strcpy(fname,"bsqi");
    sprintf(numstr, "%d", ecgsignals[i]);
    strcat(fname,numstr);
    strcat(fname,".txt");
    printf("fname is %s\n",fname);fflush(stdout);
    fid=fopen(fname,"r");
    while(fscanf(fid,"%f",&value)!=EOF)
      ecgscomparedetections[i].push_back(value);
    fclose(fid);

    printf("for ECG%d there are %d power intervals %d detection intervals\n",
	  i,ecgspower[i].size(),ecgscomparedetections[i].size() );
    fflush(stdout);
  }

  //read BPs quality indices int vectors
  //
  vector<vector<float> >bpspower(numbps);
  vector<vector<float> >bpscomparedetections(numbps);

  for(int i=0;i<numbps;++i){

    //bppower  
    //
    strcpy(fname,"powersignalnobase");
    sprintf(numstr, "%d", bpsignals[i]);
    strcat(fname,numstr);
    strcat(fname,".txt");
    printf("fname is %s\n",fname);fflush(stdout);
    fid=fopen(fname,"r");
    while(fscanf(fid,"%f",&value)!=EOF)
      bpspower[i].push_back(value);
    fclose(fid);

    //bpscomparedetections between bp signal and gqrs
    //
    /*
    strcpy(fname,"bsqi");
    sprintf(numstr, "%d", bpsignals[i]);
    strcat(fname,numstr);
    strcat(fname,".txt");
    printf("fname is %s\n",fname);fflush(stdout);
    fid=fopen(fname,"r");
    while(fscanf(fid,"%f",&value)!=EOF)
      bpscomparedetections[i].push_back(value);
    fclose(fid);
    */
    printf("for BP%d there are %d power intervals %d detection intervals\n",
	   i,bpspower[i].size(),bpscomparedetections[i].size() );
  }

  vector<int> signalonintervals;//for each interval one signal (if possible)
  selectsignalsonintervals(ecgsignals, bpsignals, ecgspower,ecgscomparedetections,
			   bpspower, bpscomparedetections, signalonintervals,
			   ecgspowerlimit, ecgscomparedetectionslimit,
			   bpspowerlimit, bpscomparedetectionslimit);


  vector<int> usedsignals;
  vector<vector<int> > intervalsforsignals;
  listintervalsforsignals(signalonintervals, usedsignals, intervalsforsignals);
  int numusedsignals=usedsignals.size();
  int samplingfreq=atoi(argv[3]);

  // for each used signal, read annotations on intervals
  char annotname[10];
  char strnum[12];
  char recplussigname[20];

  vector<vector<vector<int> > >detectionsonintervals;
  detectionsonintervals.resize(numusedsignals);//for each used signal,
  //for each interval where it is the chosen, contains the detections for that signal,
  // on that interval 


  for(int i=0;i<numusedsignals;++i){

    detectionsonintervals[i].resize(intervalsforsignals[i].size());

    //annotation file's name
    if(find(ecgsignals.begin(),ecgsignals.end(),usedsignals[i])!=ecgsignals.end())
      //signal is an ECG
      strcpy(annotname,".qrsbeat");//gqrs annotation
    else if(find(bpsignals.begin(),bpsignals.end(),usedsignals[i])!=bpsignals.end())
      //signal is a blood pressure
      strcpy(annotname,".bpbeat");//bp annotation

    strcpy(recplussigname,argv[2]);
    sprintf (strnum, "%d", usedsignals[i]); 
    strcat(recplussigname,strnum);

    //fill detectionsonintervals
    selectannotationsonintervalstxt(recplussigname, annotname, intervalsforsignals[i], 
				    lengthsecs, detectionsonintervals[i]);
  }


  //final set of annotations
  vector<int> finalannot;
  createfinalannotations(signalonintervals, usedsignals, detectionsonintervals, finalannot);

  fid=fopen("beatannotations.txt","w");
  for(int i=0;i<finalannot.size();++i)
    fprintf(fid,"%d\n",finalannot[i]);
  fclose(fid);


  return 0;
}

void createdictionary(const vector<int>&usedsignals, int maxnumsignals,
		      vector<int>&dictionary);


//list, for each signal, on which intervals it is the "chosen one"

void listintervalsforsignals(const vector<int>&signalonintervals, vector<int>&usedsignals,
			     vector<vector<int> >& intervalsforsignals){

  //signalonintervals, is the output of 'selectsignalsonintervals()'
  //usedsignals, signals that are used at least in one interval
  //intervalsforsignals, for each used signal contains the intervals where it is chosen 

  int sig,K=10; //max number of signals in the record
  int numintervals=signalonintervals.size();

  //first list used signals
  for(int i=0;i<K;++i)
    if(find(signalonintervals.begin(),signalonintervals.end(),i)!=signalonintervals.end())
      usedsignals.push_back(i);

  int numusedsignals=usedsignals.size();
  intervalsforsignals.resize(numusedsignals);

  vector<int> dictionary;
  createdictionary(usedsignals, K, dictionary);

  for(int j=0;j<numintervals;++j){
    sig=signalonintervals[j];
    intervalsforsignals[dictionary[sig] ].push_back(j);
  }

}


void createfinalannotations(const vector<int>&signalonintervals, const vector<int>&usedsignals,
			    vector<vector<vector<int> > >&detectionsonintervals,
			    vector<int>&finalannot){

  int K=10;//max num signals in record

  vector<int> dictionary;//for each used signal contains its order in used signals,
  // -1 if not used
  createdictionary(usedsignals, K, dictionary);

  int numusedsignals=usedsignals.size();
  vector<int>intervals(numusedsignals,0);//counter for the intervals order for
  //each used signal

  int sig,aux;
  int numintervals=signalonintervals.size();
  for(int i=0;i<numintervals;++i){

    sig=signalonintervals[i];
    aux=dictionary[sig];
    finalannot.insert(finalannot.end(),detectionsonintervals[aux][intervals[aux] ].begin(),
		      detectionsonintervals[aux][intervals[aux] ].end());
    intervals[aux]+=1;//order of sig interval to be used next time 
  }

  //debug
  for(int i=0;i<numusedsignals;++i)
    assert(intervals[i]==detectionsonintervals[i].size());
}


void createdictionary(const vector<int>&usedsignals, int maxnumsignals,
		      vector<int>&dictionary){

  int &K=maxnumsignals;

  dictionary.resize(K);
  for(int i=0;i<K;++i)
    dictionary[i]=-1;

  int sig;
  int numusedsignals=usedsignals.size();
  for(int i=0;i<numusedsignals;++i){
    sig=usedsignals[i];
    dictionary[sig]=i;
  }


}
